/**************************************************************
 *
 *
 * $Header:$
 *
 * $Log:$
 *
 * MODULE DESCRIPTION
 *
 *   Sample from middle layer of network, 
 *   below
 *
 *
 * $Author:$
 *
 *************************************************************/
#include <stdlib.h>
#include <math.h>
#include <mex.h>
#include "twogauss.h"
#include "threegauss.h"
#include "midsample.h"

static  double  pi2 = 2.0 * M_PI;
static  double  tiny = 10 / HUGE;


#define MAX(x,y)	(x > y ? x : y)
#define MIN(x,y)	(x < y ? x : y)
#define REAL		(0)
/************************************************************
 *
 * sample from top layer given middle layer
 *
 * x1 - State of lop layer
 * x2  - x2 - bias at lower layer
 * W12 - Weight matrix between layers
 * M - Lateral connections between layers
 * b1 - Bias at top layer
 *
 * Returns:  x1New - The new state x1
 *
 *************************************************************/
Matrix * midsample(
int		nrhs,
Matrix		*prhs[],
double		*newX1r
)
{
  Matrix	*x1, *x2, *x3, *W12, *W23,  *b2, *b3,*v1,  *v2, *v3, *nl;
  Matrix	*y = NULL, *x1r = NULL, *x2r= NULL, *ret = NULL;
  int		i = 0, j;
  int		ns, totSamp;
  double	mua, va, md, vd, mjj;
  int		n1, n2, n3, NL;

  /* Pointers to matrix data */
  double	*pW23, *pb2, *pb3,  *px1, *px2, *px3, *px2r, *px1r, *pv2, *pv3;
  double	*pRet, *py;


  x1 = prhs[i++];
  x2 = prhs[i++];
  x3 = prhs[i++];
  W12 = prhs[i++];
  W23 = prhs[i++];
  b2 = prhs[i++];
  b3 = prhs[i++];
  v1 = prhs[i++];
  v2 = prhs[i++];
  v3 = prhs[i++];
  nl = prhs[i++];

  n1 = mxGetM(x1);
  n2 = mxGetM(x2);
  n3 = mxGetM(x3);

  if (mxGetM(W12) != n2 && mxGetN(W12) + n1)
  {
    mexErrMsgTxt("topsample: W12 does not match dimensions of x1 and x2");
  }

  if (mxGetM(W23) != n3 && mxGetN(W23) != n2)
  {
    mexErrMsgTxt("topsample: W23 does not match dimensions of x2 and x3");
  }

  x1r = mxCreateFull(n1, 1, REAL);
  x2r = mxCreateFull(n2, 1, REAL);
  y = mxCreateFull(n3, 1, REAL);

  pb2 = mxGetPr(b2);
  pb3 = mxGetPr(b3);
  px1 = mxGetPr(x1);
  px2 = mxGetPr(x2);
  px3 = mxGetPr(x3);
  pv2 = mxGetPr(v2);
  pv3 = mxGetPr(v3);
  px1r = mxGetPr(x1r);
  px2r = mxGetPr(x2r);
  pW23 = mxGetPr(W23);
  NL = (int) *mxGetPr(nl);
  py = mxGetPr(y);

  ret = mxCreateFull(n2, 1, REAL);
  pRet = mxGetPr(ret);


  /* rectified top activities */
  if (NL==1)
    for (j = 0 ; j < n1 ; ++j)
      {
	px1r[j] = MAX(px1[j], 0.0);
      }
  else if (NL==3)
    for (j = 0 ; j < n1 ; ++j)
      {
	px1r[j] = MAX(px1[j], 0.0);
	px1r[j] = MIN(px1r[j], 1.0);
      }
  /* rectVect(px1r, px1, n1); */ 
  rectVect(px2r, px2, n2); /* middle layer is always NL=1 */

  for ( j = 0 ; j < n2 ; ++j)
  {
    /* top down mean and var */
    mua = dotProd(W12, j, x1r, 0) + pb2[j];
    va = *pv2;


   /* Combine top down with bottom up */
    for (i = 0 ; i < n3 ; ++i)
    {
      py[i] = px3[i] - dotProd(W23, i, x2r, 0) - pb3[i] + pW23[j*n3 + i] * px2r[j];
    }

    vd = 1 / (1/va + transDotProd(W23, j, W23, j) / *pv3);

    md = vd * (mua/va + transDotProd(y, 0, W23, j) / *pv3);
  
#ifdef DEBUG
    printf("%d %f %f %f\n", j, md, vd,  transDotProd(y, 0, W23, j));
#endif
    
    
    px2[j] =twogauss(mua,va,md,vd); /* Middle layer is always assumed NL=1 */
    
  

    px2r[j] = MAX(px2[j], 0.0);
    pRet[j] = px2[j];
  }

  return (ret);
}


void mexFunction(
int		nlhs,
Matrix		*plhs[],
int		nrhs,
Matrix		*prhs[]
)
{
  double	*retx1;
  int		i = 0;
  double	*retX1;
  static int	expectRhs = 11;
  static int	expectLhs = 1;
  char		buffer[100];

  /* Check for proper number of arguments */
  
  if (nrhs != expectRhs) 
  {
    sprintf(buffer, "midsample requires %d input arguments. Got %d",
	    expectRhs, nrhs);
    mexErrMsgTxt(buffer);
  } else if (nlhs > 1) 
  {
    sprintf(buffer, "midsample requires %d outputs arguments. Got %d",
	    expectLhs, nlhs);
    mexErrMsgTxt(buffer);
  }

  plhs[0] = midsample(nrhs, prhs, retX1);
}
